class RouteWithTooSmallCapacity(Exception):
    pass


class PaymentFailure(Exception):
    pass


class DryRun(Exception):
    pass


class PaymentTimeOut(Exception):
    pass


class TooExpensive(Exception):
    pass


class RebalanceFailure(Exception):
    pass


class RoutesExhausted(RebalanceFailure):
    pass


class RebalanceCandidatesExhausted(RebalanceFailure):
    pass


class NoRebalanceCandidates(RebalanceFailure):
    pass


class RebalancingTrialsExhausted(RebalanceFailure):
    pass


class MultichannelInboundRebalanceFailure(RebalanceFailure):
    pass


class NoRoute(Exception):
    pass


class DuplicateRoute(NoRoute):
    pass


class RPCError(Exception):
    pass


class RouterRPCError(RPCError):
    pass



#!/usr/bin/env python
import argparse
import time
import os
import sys
# readline has a desired side effect on keyword input of enabling history
import readline

from lndmanage.lib.node import LndNode
from lndmanage.lib.listchannels import ListChannels
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.recommend_nodes import RecommendNodes
from lndmanage.lib.report import Report
from lndmanage.lib.exceptions import (
    DryRun,
    PaymentTimeOut,
    TooExpensive,
    RebalanceFailure,
    RebalancingTrialsExhausted,
)
from lndmanage import settings

import logging.config
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()


def range_limited_float_type(unchecked_value):
    """
    Type function for argparse - a float within some predefined bounds

    :param: unchecked_value: float
    """
    try:
        value = float(unchecked_value)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if value < 1E-6 or value > 1:
        raise argparse.ArgumentTypeError(
            "Argument must be < " + str(1E-6) + " and > " + str(1))
    return value


def unbalanced_float(x):
    """
    Checks if the value is a valid unbalancedness between [-1 ... 1]
    """
    x = float(x)
    if x < -1.0 or x > 1.0:
        raise argparse.ArgumentTypeError(f"{x} not in range [-1.0, 1.0]")
    return x


class Parser(object):
    def __init__(self):
        # setup the command line parser
        self.parser = argparse.ArgumentParser(
            prog='lndmanage.py',
            description='Lightning network daemon channel management tool.')
        self.parser.add_argument(
            '--loglevel', default='INFO', choices=['INFO', 'DEBUG'])
        subparsers = self.parser.add_subparsers(dest='cmd')

        self.parser_status = subparsers.add_parser(
            'status', help='display node status',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        # cmd: listchannels
        self.parser_listchannels = subparsers.add_parser(
            'listchannels',
            help='lists channels with extended information '
                 '[see also subcommands with -h]',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        listchannels_subparsers = self.parser_listchannels.add_subparsers(
            dest='subcmd')

        # subcmd: listchannels rebalance
        parser_listchannels_rebalance = listchannels_subparsers.add_parser(
            'rebalance', help='displays unbalanced channels')
        parser_listchannels_rebalance.add_argument(
            '--unbalancedness', type=float,
            default=settings.UNBALANCED_CHANNEL,
            help='Unbalancedness is a way to express how balanced a '
                 'channel is, a value between [-1, 1] (a perfectly balanced '
                 'channel has a value of 0). The flag excludes channels with '
                 'an absolute unbalancedness smaller than UNBALANCEDNESS.')
        parser_listchannels_rebalance.add_argument(
            '--sort-by', default='rev_ub', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels inactive
        parser_listchannels_inactive = listchannels_subparsers.add_parser(
            'inactive', help="displays inactive channels")
        parser_listchannels_inactive.add_argument(
            '--sort-by', default='lupp', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels forwardings
        parser_listchannels_forwardings = listchannels_subparsers.add_parser(
            'forwardings',
            help="displays channels with forwarding information")
        parser_listchannels_forwardings.add_argument(
            '--from-days-ago', default=30, type=int,
            help='time interval start (days ago)')
        parser_listchannels_forwardings.add_argument(
            '--to-days-ago', default=0, type=int,
            help='time interval end (days ago)')
        parser_listchannels_forwardings.add_argument(
            '--sort-by', default='f/w', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels hygiene
        parser_listchannels_hygiene = listchannels_subparsers.add_parser(
            'hygiene',
            help="displays channels with information for channel closing")
        parser_listchannels_hygiene.add_argument(
            '--from-days-ago', default=60, type=int,
            help='time interval start (days ago)')
        parser_listchannels_hygiene.add_argument(
            '--sort-by', default='rev_nfwd', type=str,
            help='sort by column (look at description)')

        # cmd: rebalance
        self.parser_rebalance = subparsers.add_parser(
            'rebalance', help='rebalance a channel',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_rebalance.add_argument('channel', type=int,
                                           help='channel_id')
        self.parser_rebalance.add_argument(
            '--max-fee-sat', type=int, default=20,
            help='Sets the maximal fees in satoshis to be paid.')
        self.parser_rebalance.add_argument(
            '--chunksize', type=float, default=1.0,
            help='Specifies if the individual rebalance attempts should be '
                 'split into smaller relative amounts. This increases success'
                 ' rates, but also increases costs!')
        self.parser_rebalance.add_argument(
            '--max-fee-rate', type=range_limited_float_type, default=5E-5,
            help='Sets the maximal effective fee rate to be paid.'
                 ' The effective fee rate is defined by '
                 '(base_fee + amt * fee_rate) / amt.')
        self.parser_rebalance.add_argument(
            '--reckless', help='Execute action in the network.',
            action='store_true')
        self.parser_rebalance.add_argument(
            '--allow-unbalancing',
            help=f'Allow channels to get an unbalancedness'
            f' up to +-{settings.UNBALANCED_CHANNEL}.',
            action='store_true')
        self.parser_rebalance.add_argument(
            '--target', help=f'This feature is still experimental! '
            f'The unbalancedness target is between [-1, 1]. '
            f'A target of -1 leads to a maximal local balance, a target of 0 '
            f'to a 50:50 balanced channel and a target of 1 to a maximal '
            f'remote balance. Default is a target of 0.',
            type=unbalanced_float, default=0.0)
        rebalancing_strategies = ['most-affordable-first',
                                  'lowest-feerate-first', 'match-unbalanced']
        self.parser_rebalance.add_argument(
            '--strategy',
            help=f'Rebalancing strategy.',
            choices=rebalancing_strategies, type=str, default=None)

        # cmd: circle
        self.parser_circle = subparsers.add_parser(
            'circle', help='circular self-payment',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_circle.add_argument('channel_from', type=int,
                                        help='channel_from')
        self.parser_circle.add_argument('channel_to', type=int,
                                        help='channel_from')
        self.parser_circle.add_argument('amt_sat', type=int,
                                        help='amount in satoshis')
        self.parser_circle.add_argument(
            '--max-fee-sat', type=int, default=20,
            help='Sets the maximal fees in satoshis to be paid.')
        self.parser_circle.add_argument(
            '--max-fee-rate', type=range_limited_float_type, default=5E-5,
            help='Sets the maximal effective fee rate to be paid. '
                 'The effective fee rate is defined by '
                 '(base_fee + amt * fee_rate) / amt.')
        self.parser_circle.add_argument(
            '--reckless', help='Execute action in the network.',
            action='store_true')

        # cmd: recommend-nodes
        self.parser_recommend_nodes = subparsers.add_parser(
            'recommend-nodes',
            help='recommends nodes [see also subcommands with -h]',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_recommend_nodes.add_argument(
            '--show-connected', action='store_true', default=False,
            help='specifies if already connected nodes should be '
                 'removed from list')
        self.parser_recommend_nodes.add_argument(
            '--show-addresses', action='store_true', default=False,
            help='specifies if node addresses should be shown')
        parser_recommend_nodes_subparsers = \
            self.parser_recommend_nodes.add_subparsers(
                dest='subcmd')

        # TODO: put global options to the
        #  parent parser (e.g. number of nodes, sort-by flag)

        # subcmd: recommend-nodes good-old
        parser_recommend_nodes_good_old = \
            parser_recommend_nodes_subparsers.add_parser(
                'good-old',
                help='shows nodes already interacted with but no '
                     'active channels',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_good_old.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_good_old.add_argument(
            '--sort-by', default='tot', type=str,
            help="sort by column [abbreviation, e.g. 'tot']")

        # subcmd: recommend-nodes flow-analysis
        parser_recommend_nodes_flow_analysis = \
            parser_recommend_nodes_subparsers.add_parser(
                'flow-analysis', help='recommends nodes from a flow analysis',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_flow_analysis.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--forwarding-events', default=200, type=int,
            help='sets the number of forwarding events in the flow analysis')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--inwards', action='store_true',
            help='if True, inward-flowing nodes are displayed '
                 'instead of outward-flowing nodes')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--sort-by', default='weight', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # subcmd: recommend-nodes external_source
        parser_recommend_nodes_external_source = \
            parser_recommend_nodes_subparsers.add_parser(
                'external-source',
                help='recommends nodes from a given file/url',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_external_source.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_external_source.add_argument(
            '--source', type=str,
            default='https://github.com/lightningnetworkstores/'
                    'lightningnetworkstores.github.io/raw/master/sites.json',
            help='url/file to be analyzed')
        parser_recommend_nodes_external_source.add_argument(
            '--distributing-nodes', action='store_true',
            help='if True, distributing nodes are '
                 'displayed instead of the bare nodes')
        parser_recommend_nodes_external_source.add_argument(
            '--sort-by', default='cpc', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # subcmd: recommend-nodes channel-openings
        parser_recommend_nodes_channel_openings = \
            parser_recommend_nodes_subparsers.add_parser(
                'channel-openings',
                help='recommends nodes from recent channel openings',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_channel_openings.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_channel_openings.add_argument(
            '--from-days-ago', type=int,
            default=30,
            help='channel openings starting from a time frame days ago')
        parser_recommend_nodes_channel_openings.add_argument(
            '--sort-by', default='msteady', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # cmd: report
        parser_report = subparsers.add_parser(
            'report',
            help="displays reports of activity on the node")
        parser_report.add_argument(
            '--from-days-ago', default=1, type=int,
            help='time interval start (days ago)')
        parser_report.add_argument(
            '--to-days-ago', default=0, type=int,
            help='time interval end (days ago)')

    def parse_arguments(self):
        return self.parser.parse_args()

    def run_commands(self, node, args):
        # program execution
        if args.loglevel:
            # update the loglevel of the stdout handler to the user choice
            logger.handlers[0].setLevel(args.loglevel)

        if args.cmd == 'status':
            node.print_status()

        elif args.cmd == 'listchannels':
            listchannels = ListChannels(node)
            if not args.subcmd:
                listchannels.print_all_channels('rev_alias')
            if args.subcmd == 'rebalance':
                listchannels.print_channels_unbalanced(
                    args.unbalancedness, sort_string=args.sort_by)
            elif args.subcmd == 'inactive':
                listchannels.print_channels_inactive(
                    sort_string=args.sort_by)
            elif args.subcmd == 'forwardings':
                # convert time interval into unix timestamp
                time_from = time.time() - args.from_days_ago * 24 * 60 * 60
                time_to = time.time() - args.to_days_ago * 24 * 60 * 60
                logger.info(
                    f"Forwardings from {args.from_days_ago} days ago"
                    f" to {args.to_days_ago} days ago are included.")
                listchannels.print_channels_forwardings(
                    time_interval_start=time_from, time_interval_end=time_to,
                    sort_string=args.sort_by)
            elif args.subcmd == 'hygiene':
                time_from = time.time() - args.from_days_ago * 24 * 60 * 60
                logger.info(f"Channel hygiene stats is over last "
                            f"{args.from_days_ago} days.")
                listchannels.print_channels_hygiene(
                    time_interval_start=time_from, sort_string=args.sort_by)

        elif args.cmd == 'rebalance':
            if args.target:
                logger.warning("Warning: Target is set, this is still an "
                               "experimental feature.")
            rebalancer = Rebalancer(node, args.max_fee_rate, args.max_fee_sat)
            try:
                rebalancer.rebalance(
                    args.channel,
                    dry=not args.reckless,
                    chunksize=args.chunksize,
                    target=args.target,
                    allow_unbalancing=args.allow_unbalancing,
                    strategy=args.strategy)
            except TooExpensive as e:
                logger.error(f"Too expensive: {e}")
            except RebalanceFailure as e:
                logger.error(f"Rebalance failure: {e}")

        elif args.cmd == 'circle':
            rebalancer = Rebalancer(node, args.max_fee_rate, args.max_fee_sat)
            invoice_r_hash = node.get_rebalance_invoice(
                memo='circular payment')
            try:
                rebalancer.rebalance_two_channels(
                    args.channel_from, args.channel_to,
                    args.amt_sat, invoice_r_hash, args.max_fee_sat,
                    dry=not args.reckless)
            except DryRun:
                logger.info("This was just a dry run.")
            except TooExpensive:
                logger.error(
                    "Too expensive: consider to raise --max-fee-sat or "
                    "--max-fee-rate.")
            except RebalancingTrialsExhausted:
                logger.error(
                    f"Rebalancing trials exhausted (number of trials: "
                    f"{settings.REBALANCING_TRIALS}).")
            except PaymentTimeOut:
                logger.error("Payment failed because the payment timed out.")

        elif args.cmd == 'recommend-nodes':
            if not args.subcmd:
                self.parser_recommend_nodes.print_help()
                return 0

            recommend_nodes = RecommendNodes(
                node, show_connected=args.show_connected,
                show_addresses=args.show_addresses)

            if args.subcmd == 'good-old':
                recommend_nodes.print_good_old(number_of_nodes=args.nnodes,
                                               sort_by=args.sort_by)
            elif args.subcmd == 'flow-analysis':
                recommend_nodes.print_flow_analysis(
                    out_direction=(not args.inwards),
                    number_of_nodes=args.nnodes,
                    forwarding_events=args.forwarding_events,
                    sort_by=args.sort_by)
            elif args.subcmd == 'external-source':
                recommend_nodes.print_external_source(
                    args.source, distributing_nodes=args.distributing_nodes,
                    number_of_nodes=args.nnodes, sort_by=args.sort_by)
            elif args.subcmd == 'channel-openings':
                recommend_nodes.print_channel_openings(
                    from_days_ago=args.from_days_ago,
                    number_of_nodes=args.nnodes, sort_by=args.sort_by)

        elif args.cmd == 'report':
            time_from = time.time() - args.from_days_ago * 24 * 60 * 60
            time_to = time.time() - args.to_days_ago * 24 * 60 * 60
            report = Report(node, time_from, time_to)
            report.report()


def main():
    parser = Parser()

    # config.ini is expected to be in home/.lndmanage directory
    config_file = os.path.join(settings.home_dir, 'config.ini')

    # if lndmanage is run with arguments, run once
    if len(sys.argv) > 1:
        # take arguments from sys.argv
        args = parser.parse_arguments()
        node = LndNode(config_file=config_file)
        parser.run_commands(node, args)

    # otherwise enter an interactive mode
    else:
        history_file = os.path.join(settings.home_dir, "command_history")
        try:
            readline.read_history_file(history_file)
        except FileNotFoundError:
            # history will be written later
            pass

        logger.info("Running in interactive mode. "
                    "You can type 'help' or 'exit'.")
        node = LndNode(config_file=config_file)

        while True:
            try:
                user_input = input("$ lndmanage ")
            except (EOFError, KeyboardInterrupt):
                readline.write_history_file(history_file)
                logger.info("exit")
                return 0

            if not user_input or user_input in ['help', '-h', '--help']:
                parser.parser.print_help()
                continue
            elif user_input == 'exit':
                readline.write_history_file(history_file)
                return 0

            args_list = user_input.split(" ")
            try:
                # need to run with parse_known_args to get an exception
                args = parser.parser.parse_args(args_list)
                parser.run_commands(node, args)
            except:
                logger.exception("Exception encountered.")
                continue


if __name__ == '__main__':
    main()



"""
Tests for circular self-payments.
"""
import time
from unittest import TestCase

from lnregtest.lib.network import RegtestNetwork

from lndmanage.lib.node import LndNode
from lndmanage.lib.listchannels import ListChannels
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.exceptions import (
    RebalanceFailure,
    TooExpensive,
    DryRun,
    RebalancingTrialsExhausted,
    NoRoute,
    DuplicateRoute,
)
from lndmanage import settings

from test.testing_common import (
    bin_dir,
    test_data_dir,
    test_graphs_paths,
    lndmanage_home,
    SLEEP_SEC_AFTER_REBALANCING
)

import logging.config
settings.set_lndmanage_home_dir(lndmanage_home)
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger.handlers[0].setLevel(logging.DEBUG)


class CircleTest(TestCase):

    network_definition = None

    def setUp(self):
        if self.network_definition is None:
            self.skipTest("This class doesn't represent a real test case.")
            raise NotImplementedError("A network definition needs to be given.")

        self.testnet = RegtestNetwork(
            binary_folder=bin_dir,
            network_definition_location=self.network_definition,
            nodedata_folder=test_data_dir,
            node_limit='H',
            from_scratch=True
        )
        self.testnet.run_nocleanup()

        master_node_data_dir = self.testnet.master_node.lnd_data_dir
        master_node_port = self.testnet.master_node.grpc_port
        self.master_node_networkinfo = self.testnet.master_node.getnetworkinfo()

        self.lndnode = LndNode(
            lnd_home=master_node_data_dir,
            lnd_host='localhost:' + str(master_node_port),
            regtest=True
        )

        self.rebalancer = Rebalancer(
            self.lndnode,
            max_effective_fee_rate=50,
            budget_sat=20
        )
        self.graph_test()

    def tearDown(self):
        self.testnet.cleanup()

    def graph_test(self):
        raise NotImplementedError

    def circle_and_check(self, rebalancer, channel_number_send,
                         channel_number_receive, amount_sat,
                         expected_fees_msat, dry=False):
        """
        Helper function for testing a circular payment.

        :param rebalancer:
        :type rebalancer: lndmanage.lib.rebalance.Rebalancer

        :param channel_number_send: channel whose local balance is decreased
        :type channel_number_send: int

        :param channel_number_receive: channel whose local balance is increased
        :type channel_number_receive: int

        :param amount_sat: amount in satoshi to rebalance
        :type amount_sat: int

        :param expected_fees_msat: expected fees in millisatoshi for
            the rebalance
        :type expected_fees_msat: int

        :param dry: if it should be a dry run
        :type dry: bool
        """
        # setup
        graph_before = self.testnet.assemble_graph()
        channel_id_send = self.testnet.channel_mapping[
            channel_number_send]['channel_id']
        channel_id_receive = self.testnet.channel_mapping[
            channel_number_receive]['channel_id']
        invoice_r_hash = self.lndnode.get_invoice(amount_sat, '')

        # exercise
        try:
            fees_msat = rebalancer.rebalance_two_channels(
                channel_id_send,
                channel_id_receive,
                amount_sat,
                invoice_r_hash,
                rebalancer.budget_sat,
                dry=dry
            )
            time.sleep(SLEEP_SEC_AFTER_REBALANCING)
        except Exception as e:
            raise e

        # check
        graph_after = self.testnet.assemble_graph()
        channel_data_send_before = graph_before['A'][channel_number_send]
        channel_data_receive_before = graph_before['A'][channel_number_receive]
        channel_data_send_after = graph_after['A'][channel_number_send]
        channel_data_receive_after = graph_after['A'][channel_number_receive]
        listchannels = ListChannels(self.lndnode)
        listchannels.print_all_channels('rev_alias')

        # test that the fees are correct
        self.assertEqual(fees_msat, expected_fees_msat)
        # test if sending channel's remote balance has increased correctly
        self.assertEqual(
            amount_sat,
            channel_data_send_after['remote_balance'] -
            channel_data_send_before['remote_balance'] -
            int(expected_fees_msat // 1000),
            "Sending local balance is wrong"
        )
        # test if receiving channel's local balance has increased correctly
        self.assertEqual(
            amount_sat,
            channel_data_receive_after['local_balance'] -
            channel_data_receive_before['local_balance'],
            "Receiving local balance is wrong"
        )


class TestCircleLiquid(CircleTest):
    network_definition = test_graphs_paths['star_ring_3_liquid']

    def graph_test(self):
        # assert some basic properties of the graph
        self.assertEqual(4, self.master_node_networkinfo['num_nodes'])
        self.assertEqual(6, self.master_node_networkinfo['num_channels'])

    def test_circle_success_1_2(self):
        """
        Test successful rebalance from channel 1 to channel 2.
        """
        channel_number_from = 1
        channel_number_to = 2
        amount_sat = 10000
        expected_fees_msat = 43

        self.circle_and_check(
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_success_1_6(self):
        """
        Test successful rebalance from channel 1 to channel 6.
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 10000
        expected_fees_msat = 33

        self.circle_and_check(
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_6_1_fail_rebalance_failure_no_funds(self):
        """
        Test expected failure for channel 6 to channel 1, where channel 6
        doesn't have funds.
        """
        channel_number_from = 6
        channel_number_to = 1
        amount_sat = 10000
        expected_fees_msat = 33

        self.assertRaises(
            RebalanceFailure,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_fail_budget_too_expensive(self):
        """
        Test expected failure where rebalance uses more than the fee budget.
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 10000
        expected_fees_msat = 33
        self.rebalancer.budget_sat = 0

        self.assertRaises(
            TooExpensive,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_fail_max_fee_rate_too_expensive(self):
        """
        Test expected failure where rebalance is more expensive than
        the desired maximal fee rate.
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 10000
        expected_fees_msat = 33
        self.rebalancer.max_effective_fee_rate = 0

        self.assertRaises(
            TooExpensive,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_fail_first_hop_insufficient_funds(self):
        """
        Tests failure when our channel fails because we didn't take into
        account channel reserve. (Failing channel 1: 117:1:0)
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 1000000
        expected_fees_msat = 33

        self.assertRaises(
            RebalanceFailure,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_success_channel_reserve(self):
        """
        Test for a maximal amount circular payment.
        """
        channel_number_from = 1
        channel_number_to = 6
        local_balance = 1000000
        # take into account 1% channel reserve
        amount_sat = int(local_balance - 0.01 * local_balance)
        # need to also subtract commitment fees
        amount_sat -= 9050
        # need to also subtract fees, then error message changes
        amount_sat -= 3
        # extra to make it succeed
        amount_sat -= 2150

        expected_fees_msat = 2938

        self.circle_and_check(
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_fail_rebalance_dry(self):
        """
        Test if dry run exception is raised.
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 10000
        expected_fees_msat = 33

        self.assertRaises(
            DryRun,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
            dry=True
        )


class TestCircleIlliquid(CircleTest):

    network_definition = test_graphs_paths['star_ring_4_illiquid']

    def graph_test(self):
        self.assertEqual(5, self.master_node_networkinfo['num_nodes'])
        self.assertEqual(10, self.master_node_networkinfo['num_channels'])

    def test_circle_fail_2_3_no_route(self):
        """
        Test if NoRoute is raised.
        """
        channel_number_from = 2
        channel_number_to = 3
        amount_sat = 500000
        expected_fees_msat = None

        self.assertRaises(
            NoRoute,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_1_2_fail_max_trials_exhausted(self):
        """
        Test if RebalancingTrialsExhausted is raised.
        """
        channel_number_from = 1
        channel_number_to = 2
        amount_sat = 190950
        expected_fees_msat = None
        settings.REBALANCING_TRIALS = 1

        self.assertRaises(
            RebalancingTrialsExhausted,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat
        )

    # def test_circle_1_2_success_multi_trial(self):
    #     """
    #     Test if RebalancingTrialsExhausted is raised.
    #     """
    #     channel_number_from = 1
    #     channel_number_to = 2
    #     amount_sat = 200000
    #     expected_fees_msat = 1207

    #     self.circle_and_check(
    #         self.rebalancer,
    #         channel_number_from,
    #         channel_number_to,
    #         amount_sat,
    #         expected_fees_msat)

    def test_circle_1_2_fail_no_route_multi_trials(self):
        """
        Test if NoRoute is raised.
        """
        channel_number_from = 1
        channel_number_to = 2
        amount_sat = 450000
        expected_fees_msat = None

        self.assertRaises(
            RebalancingTrialsExhausted,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat
        )




import logging
from collections import OrderedDict, defaultdict

import numpy as np

from lndmanage.lib.node import LndNode
from lndmanage import settings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

np.warnings.filterwarnings('ignore')

# nearest neighbor weight for flow-analysis ~1/(avg. degree)
NEIGHBOR_WEIGHT = 0.1
# nearest neighbor weight for flow-analysis ~1/(avg. degree)^2
NEXT_NEIGHBOR_WEIGHT = 0.02


def nan_to_zero(number):
    """
    Converts float('nan') to 0
    :param number: float
    :return: float
    """
    if number is np.nan or number != number:
        return 0.0
    else:
        return number


class ForwardingAnalyzer(object):
    """
    Analyzes forwardings for single channels.
    """
    def __init__(self, node):
        self.node = node
        self.forwarding_events = self.node.get_forwarding_events()
        self.channels = {}
        self.total_forwarding_amount_sat = 0
        self.total_forwarding_fees_msat = 0
        self.forwardings = 0
        self.cumulative_effective_fee = 0
        self.timestamp_first_send = 1E10  # somewhere in future
        self.timestamp_last_send = 0  # at beginning of time
        self.max_time_interval = None

    def initialize_forwarding_data(self, time_start, time_end):
        """
        Initializes the channel statistics objects with data from
        the forwardings.

        :param time_start: time interval start, unix timestamp
        :param time_end: time interval end, unix timestamp
        """
        for f in self.forwarding_events:
            if time_start < f['timestamp'] < time_end:
                # make a dictionary entry for unknown channels
                channel_id_in = f['chan_id_in']
                channel_id_out = f['chan_id_out']

                if channel_id_in not in self.channels.keys():
                    self.channels[channel_id_in] = ChannelStatistics(
                        channel_id_in)
                if channel_id_out not in self.channels.keys():
                    self.channels[channel_id_out] = ChannelStatistics(
                        channel_id_out)

                self.channels[channel_id_in].inward_forwardings.append(
                    f['amt_in'])
                self.channels[channel_id_out].outward_forwardings.append(
                    f['amt_out'])
                self.channels[channel_id_out].absolute_fees.append(
                    f['fee_msat'])
                self.channels[channel_id_out].effective_fees.append(
                    f['effective_fee'])
                self.channels[channel_id_out].timestamps.append(
                    f['timestamp'])

                self.total_forwarding_amount_sat += f['amt_in']
                self.total_forwarding_fees_msat += f['fee_msat']
                self.forwardings += 1
                self.cumulative_effective_fee += f['effective_fee']

    def get_forwarding_statistics_channels(self):
        """
        Prepares the forwarding statistics for each channel.
        :return: dict: statistics with channel_id as keys
        """
        channel_statistics = {}

        for k, c in self.channels.items():

            try:
                timestamp_first_send = min(c.timestamps)
                if self.timestamp_first_send > timestamp_first_send:
                    self.timestamp_first_send = timestamp_first_send
            except ValueError:
                pass
            try:
                timestamp_last_send = max(c.timestamps)
                if self.timestamp_last_send < timestamp_last_send:
                    self.timestamp_last_send = timestamp_last_send
            except ValueError:
                pass

            channel_statistics[k] = {
                'effective_fee': c.effective_fee(),
                'fees_total': c.fees_total(),
                'flow_direction': c.flow_direction(),
                'mean_forwarding_in': c.mean_forwarding_in(),
                'mean_forwarding_out': c.mean_forwarding_out(),
                'median_forwarding_in': c.median_forwarding_in(),
                'median_forwarding_out': c.median_forwarding_out(),
                'number_forwardings': c.number_forwardings(),
                'largest_forwarding_amount_in':
                    c.largest_forwarding_amount_in(),
                'largest_forwarding_amount_out':
                    c.largest_forwarding_amount_out(),
                'total_forwarding_in': c.total_forwarding_in(),
                'total_forwarding_out': c.total_forwarding_out(),
            }

        # determine the time interval starting with the first forwarding
        # to the last forwarding in the analyzed time interval determined
        # by time_start and time_end
        self.max_time_interval = \
            (self.timestamp_last_send - self.timestamp_first_send) \
            / (24 * 60 * 60)

        return channel_statistics

    def get_forwarding_statistics_nodes(self, sort_by='total_forwarding'):
        """
        Calculates forwarding statistics based on single channels for
        their nodes.

        :param sort_by: str, abbreviation for the dict key
        :return:
        """
        channel_statistics = self.get_forwarding_statistics_channels()
        closed_channels = self.node.get_closed_channels()
        open_channels = self.node.get_open_channels()
        logger.debug(f"Number of channels with known forwardings: "
                     f"{len(closed_channels) + len(open_channels)} "
                     f"(thereof {len(closed_channels)} closed channels).")

        node_statistics = OrderedDict()

        # go through channel statistics and calculate node statistics
        for k, n in channel_statistics.items():
            # historic node data can be outdated, so we need to take care
            # that the remote pub key is known,
            # otherwise it is useless information

            channel_data = open_channels.get(k, None)
            if not channel_data:
                channel_data = closed_channels.get(k, None)

            if channel_data:
                remote_pubkey = channel_data['remote_pubkey']
                if remote_pubkey not in node_statistics.keys():
                    node_statistics[remote_pubkey] = {
                        'total_forwarding_in': n['total_forwarding_in'],
                        'total_forwarding_out': n['total_forwarding_out'],
                        'total_forwarding':
                            n['total_forwarding_in'] +
                            n['total_forwarding_out'],
                    }
                else:
                    node_statistics[remote_pubkey]['total_forwarding_in'] \
                        += n['total_forwarding_in']
                    node_statistics[remote_pubkey]['total_forwarding_out'] \
                        += n['total_forwarding_out']
                    node_statistics[remote_pubkey]['total_forwarding'] \
                        += (n['total_forwarding_in'] +
                            n['total_forwarding_out'])

        for k, n in node_statistics.items():
            tot_in = n['total_forwarding_in']
            tot_out = n['total_forwarding_out']
            node_statistics[k]['flow_direction'] = \
                -((float(tot_in) / (tot_in + tot_out)) - 0.5) / 0.5

        sorted_dict = OrderedDict(
            sorted(node_statistics.items(), key=lambda x: -x[1][sort_by]))

        return sorted_dict

    def simple_flow_analysis(self, last_forwardings_to_analyze=100):
        """
        Takes each forwarding event and determines the set of incoming nodes
        (up to second nearest neighbors) and outgoing nodes
        (up to second nearest neighbors) and does a frequency analysis of both
        sets, assigning a probability that a certain node was involved in
        the forwarding process.

        These probabilities are then used to take the difference between
        both, excluding effectively the large hubs. The sets can be weighted
        by a quantity of the forwarding process, e.g. the forwarding fee.

        Two lists are returned, the rest of the incoming and the rest of the
        outgoing node sets.

        :param last_forwardings_to_analyze:
            number of last forwardings to be analyzed
        :type last_forwardings_to_analyze: int

        :return: sending list, receiving list
        """
        # TODO: refine with channel capacities
        # TODO: refine with update time
        # TODO: refine with route calculation
        # TODO: fine tune NEIGHBOR_WEIGHT, NEXT_NEIGHBOR_WEIGHT

        logger.info("Doing simple forwarding analysis.")

        # initialize node score dictionaries
        total_incoming_neighbors = defaultdict(float)
        total_outgoing_neighbors = defaultdict(float)

        logger.info(f"Total forwarding events found: "
                    f"{len(self.forwarding_events)}.")
        logger.info(f"Carrying out flow analysis for last "
                    f"{last_forwardings_to_analyze} forwarding events.")
        number_progress_report = last_forwardings_to_analyze // 10

        meaningful_outward_forwardings = 0
        for nf, f in enumerate(
                self.forwarding_events[-last_forwardings_to_analyze:]):

            # report progress
            if nf % number_progress_report == 0:
                logger.info(
                    f"Analysis progress: "
                    f"{100 * float(nf) / last_forwardings_to_analyze}%")

            chan_id_in = f['chan_id_in']
            chan_id_out = f['chan_id_out']

            edge_data_in = self.node.network.edges.get(chan_id_in, None)
            edge_data_out = self.node.network.edges.get(chan_id_out, None)

            if edge_data_in is not None and edge_data_out is not None:
                # determine incoming and outgoing node pub keys
                incoming_node_pub_key = edge_data_in['node1_pub'] \
                    if edge_data_in['node1_pub'] != self.node.pub_key \
                    else edge_data_in['node2_pub']
                outgoing_node_pub_key = edge_data_out['node1_pub'] \
                    if edge_data_out['node1_pub'] != self.node.pub_key \
                    else edge_data_out['node2_pub']

                # nodes involved in the forwarding process should be excluded
                excluded_nodes = [self.node.pub_key, incoming_node_pub_key,
                    outgoing_node_pub_key]

                # determine all the nearest and second nearest
                # neighbors of the incoming/outgoing nodes,
                # they may appear more than once
                incoming_neighbors = self.__determine_joined_neighbors(
                    incoming_node_pub_key, excluded_nodes=excluded_nodes)
                outgoing_neighbors = self.__determine_joined_neighbors(
                    outgoing_node_pub_key, excluded_nodes=excluded_nodes)

                # do a symmetric difference of node sets with weights
                symmetric_difference_weights = self.__symmetric_difference(
                    incoming_neighbors, outgoing_neighbors)

                final_outgoing_nodes = self.__filter_nodes(
                    symmetric_difference_weights,
                    return_positive_weights=True)
                final_incoming_nodes = self.__filter_nodes(
                    symmetric_difference_weights,
                    return_positive_weights=False)

                # normalize the weights
                normalized_incoming = self.__normalize_neighbors(
                    final_incoming_nodes)
                normalized_outgoing = self.__normalize_neighbors(
                    final_outgoing_nodes)

                # set weight for each forwarding event
                weight = 1
                # alternatively:
                # weight = f['amt_in']
                # weight = f['fee_msat']

                for n, nv in normalized_incoming.items():
                    total_incoming_neighbors[n] += nv * weight

                # If we know, that the outward hop already reached the
                # target of the payment, we don't want to add the neighbors
                # of the outgoing node to the total statistics.
                # We know if the next hop was the final one of the payment by
                # testing whether the forwarding amount in msat has remainder
                # zero when divided by 1000 or not, provided the sent amount
                # was larger than 1E6 msat.
                if (f['amt_out_msat'] % 1000):
                    meaningful_outward_forwardings += 1
                    logger.debug(
                        f"Forwarding was not last hop: {f['amt_out_msat']}, "
                        f"chan_id_out: {chan_id_out}")
                    for n, nv in normalized_outgoing.items():
                        total_outgoing_neighbors[n] += nv * weight
        logger.info(f"Could use {meaningful_outward_forwardings} "
                    f"forwardings to estimate targets of payments.")

        # sort according to weights
        total_incoming_node_dict = self.__weighted_neighbors_to_sorted_dict(
            total_incoming_neighbors)
        total_outgoing_node_dict = self.__weighted_neighbors_to_sorted_dict(
            total_outgoing_neighbors)

        return total_incoming_node_dict, total_outgoing_node_dict

    @staticmethod
    def __weighted_neighbors_to_sorted_dict(node_dict):
        """
        Converts a node weight dictionary to a sorted list of
        nodes with weights.

        :param node_dict: dict with key-value pairs of node_pub_key and weight
        :return: sorted list
        """
        sorted_nodes_dict = OrderedDict()
        node_list = [(n, nv) for n, nv in node_dict.items()]
        node_list_sorted = sorted(node_list, key=lambda x: x[1], reverse=True)
        for n, nv in node_list_sorted:
            sorted_nodes_dict[n] = {'weight': nv}
        return sorted_nodes_dict

    def __determine_joined_neighbors(self, node_pub_key, excluded_nodes):
        """
        Determines the joined set of nearest and second neighbors and assigns
        a weight to every node dependent how often they appear.

        :param node_pub_key: str, public key of the home node
        :param excluded_nodes: list of str, public keys of excluded nodes
                                            in analysis
        :return: dict, keys: node_pub_keys, values: weights
        """

        neighbors = list(self.node.network.neighbors(node_pub_key))
        second_neighbors = list(
            self.node.network.second_neighbors(node_pub_key))

        # determine neighbor node_weights
        neighbor_weights = self.__analyze_neighbors(
            neighbors, excluded_nodes=excluded_nodes, weight=NEIGHBOR_WEIGHT)
        second_neighbor_weights = self.__analyze_neighbors(
            second_neighbors, excluded_nodes=excluded_nodes,
            weight=NEXT_NEIGHBOR_WEIGHT)

        # combine nearest and second nearest neighbor node weights
        joined_neighbors = self.__join_neighbors(
            neighbor_weights, second_neighbor_weights)

        return joined_neighbors

    @staticmethod
    def __normalize_neighbors(neighbors_weight_dict):
        """
        Normalizes the weights of the neighbors to the total weight
        of all neighbors.

        :param neighbors_weight_dict: dict, keys: node_pub_keys,
                                            values: weights
        :return: dict, keys: node_pub_keys, values: normalized weights
        """
        total_weight = 0
        normalized_dict = {}
        # determine total weight
        for n, nv in neighbors_weight_dict.items():
            total_weight += nv
        # normalize
        for n, nv in neighbors_weight_dict.items():
            normalized_dict[n] = float(nv) / total_weight

        return normalized_dict

    @staticmethod
    def __analyze_neighbors(neighbors, excluded_nodes, weight):
        """
        Analyzes a node dict for the frequency of nodes and gives them
        a weight. An upper bound of the weights is set.

        :param neighbors: list of node_pub_keys
        :param excluded_nodes: excluded node_pub_keys for analysis
        :param weight: float, weight for each individual appearance of a node
        :return: dict, keys: node_pub_keys, values: node weights
        """
        node_weights = {}

        for n in neighbors:
            if n not in excluded_nodes:
                if n in node_weights.keys():
                    node_weights[n] = min(node_weights[n] + weight, 1.0)
                else:
                    node_weights[n] = weight

        return node_weights

    @staticmethod
    def __join_neighbors(first_neighbor_dict, second_neighbor_dict):
        """
        Joins two node weight dicts together.
        :param first_neighbor_dict: dict, keys: node_pub_keys,
                                          values: node weights
        :param second_neighbor_dict: dict, keys: node_pub_keys,
                                           values: node weights
        :return: dict, keys: node_pub_keys, values: node weights
        """
        # make a copy of the first node weight dict
        joined_neighbor_dict = dict(first_neighbor_dict)

        # add all the nodes from the second dict
        for n, v in second_neighbor_dict.items():
            if n in joined_neighbor_dict:
                joined_neighbor_dict[n] = min(
                    1, joined_neighbor_dict[n] + second_neighbor_dict[n])
            else:
                joined_neighbor_dict[n] = second_neighbor_dict[n]

        return joined_neighbor_dict

    @staticmethod
    def __symmetric_difference(first_neighbors_dict, second_neighbors_dict):
        """
        Calculates the difference of weights of first and second node dicts,
        doing also a symmetric difference between the sets of nodes.

        :param first_neighbors_dict: dict, keys: node_pub_keys,
                                           values: node weights
        :param second_neighbors_dict: dict, keys: node_pub_keys,
                                            values: node weights
        :return: dict, keys: node_pub_keys, values: node weights
        """
        first_nodes = set(first_neighbors_dict.keys())
        second_nodes = set(second_neighbors_dict.keys())

        nodes_intersection = first_nodes.intersection(second_nodes)

        unique_first_nodes = first_nodes - nodes_intersection
        unique_second_nodes = second_nodes - nodes_intersection

        diff_neighbor_dict = {}
        for n in unique_first_nodes:
            diff_neighbor_dict[n] = first_neighbors_dict[n]

        for n in unique_second_nodes:
            diff_neighbor_dict[n] = second_neighbors_dict[n]

        for n in nodes_intersection:
            delta = first_neighbors_dict[n] - second_neighbors_dict[n]
            if delta != 0.0:
                diff_neighbor_dict[n] = delta

        return diff_neighbor_dict

    @staticmethod
    def __filter_nodes(node_weights, return_positive_weights=True):
        """
        Filters out nodes with positive or negative weights and
        takes the absolute.

        :param node_weights: dict
        :param return_positive_weights: bool, if True returns nodes with
                                              positive weights,
                                              if False negative weights
        :return:
        """
        new_node_weights = {}
        for n, nv in node_weights.items():
            if nv > 0 and return_positive_weights:
                new_node_weights[n] = nv
            elif nv < 0 and not return_positive_weights:
                new_node_weights[n] = -nv
        return new_node_weights


class ChannelStatistics(object):
    """
    Functionality to analyze the forwardings of a single channel.
    """
    def __init__(self, channel_id):
        self.channel_id = channel_id

        self.inward_forwardings = []
        self.outward_forwardings = []

        self.timestamps = []
        self.absolute_fees = []
        self.effective_fees = []

    def total_forwarding_in(self):
        return sum(self.inward_forwardings)

    def total_forwarding_out(self):
        return sum(self.outward_forwardings)

    def mean_forwarding_in(self):
        return np.mean(self.inward_forwardings)

    def mean_forwarding_out(self):
        return np.mean(self.outward_forwardings)

    def median_forwarding_in(self):
        return np.median(self.inward_forwardings)

    def median_forwarding_out(self):
        return np.median(self.outward_forwardings)

    def fees_total(self):
        return sum(self.absolute_fees)

    def effective_fee(self):
        return np.mean(self.effective_fees)

    def largest_forwarding_amount_out(self):
        return max(self.outward_forwardings, default=float('nan'))

    def largest_forwarding_amount_in(self):
        return max(self.inward_forwardings, default=float('nan'))

    def flow_direction(self):
        total_in = self.total_forwarding_in()
        total_out = self.total_forwarding_out()
        return -((float(total_in) / (total_in + total_out)) - 0.5) / 0.5

    def number_forwardings(self):
        return len(self.inward_forwardings) + len(self.outward_forwardings)


def get_forwarding_statistics_channels(node, time_interval_start,
                                       time_interval_end):
    """
    Joins data from listchannels and fwdinghistory to have a extended
    information about a channel.

    :param node: :class:`lib.node.Node`
    :param time_interval_start: unix timestamp
    :param time_interval_end: unix timestamp
    :return: dict of channel information with channel_id as keys
    """
    forwarding_analyzer = ForwardingAnalyzer(node)
    forwarding_analyzer.initialize_forwarding_data(
        time_interval_start, time_interval_end)

    # dict with channel_id keys
    statistics = forwarding_analyzer.get_forwarding_statistics_channels()
    logger.debug(f"Time interval (between first and last forwarding) is "
                 f"{forwarding_analyzer.max_time_interval:6.2f} days.")
    # join the two data sets:
    channels = node.get_unbalanced_channels(unbalancedness_greater_than=0.0)

    # TODO: improve this code, don't repeat
    for k, c in channels.items():
        try:  # channel forwarding statistics exists
            chan_stats = statistics[c['chan_id']]
            c['bandwidth_demand'] = max(
                nan_to_zero(chan_stats['mean_forwarding_in']),
                nan_to_zero(chan_stats['mean_forwarding_out'])
            ) / c['capacity']
            c['fees_total'] = chan_stats['fees_total']
            # time interval may be zero, to avoid zero division, replace by NaN
            try:
                c['fees_total_per_week'] = chan_stats['fees_total'] \
                    / (forwarding_analyzer.max_time_interval / 7)
            except ZeroDivisionError:
                c['fees_total_per_week'] = float('nan')
            c['flow_direction'] = chan_stats['flow_direction']
            c['median_forwarding_in'] = chan_stats['median_forwarding_in']
            c['median_forwarding_out'] = chan_stats['median_forwarding_out']
            c['mean_forwarding_in'] = chan_stats['mean_forwarding_in']
            c['mean_forwarding_out'] = chan_stats['mean_forwarding_out']
            c['number_forwardings'] = chan_stats['number_forwardings']
            c['largest_forwarding_amount_in'] = \
                chan_stats['largest_forwarding_amount_in']
            c['largest_forwarding_amount_out'] = \
                chan_stats['largest_forwarding_amount_out']
            c['total_forwarding_in'] = chan_stats['total_forwarding_in']
            c['total_forwarding_out'] = chan_stats['total_forwarding_out']

            # action required if flow same direction as unbalancedness
            # or bandwidth demand too high

            # TODO: refine 'action_required' by better metric
            if c['unbalancedness'] * c['flow_direction'] > 0 and abs(
                    c['unbalancedness']) > settings.UNBALANCED_CHANNEL:
                c['action_required'] = True
            else:
                c['action_required'] = False
            if c['bandwidth_demand'] > 0.5:
                c['action_required'] = True

        except KeyError:  # no forwarding statistics on channel is available
            c['bandwidth_demand'] = 0
            c['fees_total'] = 0
            c['fees_total_per_week'] = 0
            c['flow_direction'] = float('nan')
            c['median_forwarding_out'] = float('nan')
            c['median_forwarding_in'] = float('nan')
            c['mean_forwarding_out'] = float('nan')
            c['mean_forwarding_in'] = float('nan')
            c['number_forwardings'] = 0
            c['largest_forwarding_amount_in'] = float('nan')
            c['largest_forwarding_amount_out'] = float('nan')
            c['total_forwarding_in'] = float('nan')
            c['total_forwarding_out'] = float('nan')
            if abs(c['unbalancedness']) > settings.UNBALANCED_CHANNEL:
                c['action_required'] = True
            else:
                c['action_required'] = False
            if c['bandwidth_demand'] > 0.5:
                c['action_required'] = True
    return channels


if __name__ == '__main__':
    import time
    import logging.config
    logging.config.dictConfig(settings.logger_config)
    logger = logging.getLogger()

    nd = LndNode()
    fa = ForwardingAnalyzer(nd)
    fa.initialize_forwarding_data(time_start=0, time_end=time.time())
    print(fa.simple_flow_analysis())


import logging
import math

from lndmanage.lib.routing import Router
from lndmanage.lib.exceptions import (
    RebalanceFailure,
    NoRoute,
    NoRebalanceCandidates,
    RebalanceCandidatesExhausted,
    RebalancingTrialsExhausted,
    DryRun,
    PaymentTimeOut,
    TooExpensive,
    DuplicateRoute,
    MultichannelInboundRebalanceFailure,
)
from lndmanage import settings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Rebalancer(object):
    """
    Implements methods of rebalancing.

    A max_effective_fee_rate can be set, which limits the fee rate paid for
    a rebalance. The effective fee rate is (base_fee + fee_rate * amt)/amt.
    A fee cap can be defined by budget_sat. Individual and total rebalancing
    are not allowed to go over this amount.
    """
    def __init__(self, node, max_effective_fee_rate, budget_sat):
        """
        :param node: node instance
        :type node: lndmanage.lib.node.LndNode

        :param max_effective_fee_rate: maximum effective fee rate paid
        :type max_effective_fee_rate: float

        :param budget_sat: rebalancing budget
        :type budget_sat: int
        """
        self.node = node
        self.channel_list = {}
        self.router = Router(self.node)
        self.router.channel_rater.add_bad_node(self.node.pub_key)
        self.max_effective_fee_rate = max_effective_fee_rate
        self.budget_sat = budget_sat

    def rebalance_two_channels(self, channel_id_from, channel_id_to, amt_sat,
                               invoice_r_hash, budget_sat, dry=False):
        """
        Rebalances from channel_id_from to channel_id_to with an amount of
        amt_sat.

        A prior created `invoice_r_hash` has to be given. The budget_sat sets
        the maxmimum fees in sat that will be paid. A dry run can be done.

        :param channel_id_from: channel sending
        :type channel_id_from: int

        :param channel_id_to: channel receiving
        :type channel_id_to: int

        :param amt_sat: amount to be sent in sat
        :type amt_sat: int

        :param invoice_r_hash: payment hash
        :type invoice_r_hash: bytearray

        :param budget_sat: budget for the rebalance
        :type budget_sat: int

        :param dry: specifies if dry run
        :type dry: bool

        :return: total fees for the whole rebalance in msat
        :rtype: int

        :raises RebalancingTrialsExhausted:
            too many attempts to rebalance channels
        :raises DuplicateRoute: the same route was already tried
        :raises NoRoute: no route was found from source to destination
        :raises TooExpensive: the circular payment would exceed the fee limits
        :raises PaymentTimeOut: the payment timed out
        :raises DryRun: attempt was just a dry run
        """
        amt_msat = amt_sat * 1000
        previous_route_channel_hops = []

        count = 0
        while True:
            # only attempt a fixed number of times
            count += 1
            if count > settings.REBALANCING_TRIALS:
                raise RebalancingTrialsExhausted

            # method is set to external-mc to use mission control based
            # pathfinding
            routes = self.router.get_routes_for_rebalancing(
                channel_id_from, channel_id_to, amt_msat, method='external-mc')

            if len(routes) == 0:
                raise NoRoute
            else:
                # take only the first route from routes
                r = routes[0]

            if previous_route_channel_hops == r.channel_hops:
                raise DuplicateRoute("Have tried this route already.")

            logger.info(
                f"Next route: total fee: {r.total_fee_msat / 1000:3.3f} sat, "
                f"fee rate: {r.total_fee_msat / r.total_amt_msat:1.6f}, "
                f"hops: {len(r.channel_hops)}")
            logger.info(f"   Channel hops: {r.channel_hops}")

            rate = r.total_fee_msat / r.total_amt_msat
            if rate > self.max_effective_fee_rate:
                logger.info(
                    f"   Channel is too expensive. Rate: {rate:.6f}, "
                    f"requested max rate: {self.max_effective_fee_rate:.6f}")
                raise TooExpensive

            if r.total_fee_msat > budget_sat * 1000:
                logger.info(
                    f"   Channel is too expensive. Fee: {r.total_fee_msat:.6f}"
                    f" msat, requested max fee: {budget_sat:.6f} msat")
                raise TooExpensive

            if not dry:
                try:
                    result = self.node.send_to_route(r, invoice_r_hash)
                except PaymentTimeOut:
                    # TODO: handle payment timeout properly
                    raise PaymentTimeOut

                previous_route_channel_hops = r.channel_hops
                logger.debug(f"Payment error: {result.payment_error}.")

                if result.payment_error:
                    # determine the channel/node that reported a failure
                    reporting_channel_id = self.node.handle_payment_error(
                        result.payment_error)

                    if reporting_channel_id:
                        try:
                            index_failed_channel = \
                                r.channel_hops.index(reporting_channel_id)
                            failed_channel_id = \
                                r.channel_hops[index_failed_channel]
                        except ValueError:
                            logger.error(
                                "Failed channel not even in list of channel "
                                "hops (lnd issue?).")
                            continue
                        # check if a failed channel was our own, which should,
                        # in principle, not happen
                        if failed_channel_id in [channel_id_from,
                                                 channel_id_to]:
                            raise RebalanceFailure(
                                f"Own channel failed. Something is wrong. "
                                f"This is likely due to a wrong accounting "
                                f"for the channel reserve and will be fixed "
                                f"in the future. "
                                f"Try with smaller absolute target. "
                                f"Failing channel: {failed_channel_id}")

                        # determine the nodes involved in the channel
                        failed_channel_source = r.node_hops[
                            index_failed_channel]
                        failed_channel_target = r.node_hops[
                            index_failed_channel + 1]
                        logger.info(f"   Failed channel: {failed_channel_id}")
                        logger.debug(
                            f"   Failed channel between nodes "
                            f"{failed_channel_source} and "
                            f"{failed_channel_target}")
                        logger.debug(f"    Node hops {r.node_hops}")
                        logger.debug(f"    Channel hops {r.channel_hops}")

                        # remember the bad channel for next routing
                        self.router.channel_rater.add_bad_channel(
                            failed_channel_id, failed_channel_source,
                            failed_channel_target)

                    else:  # usually the case of UnknownNextPeer
                        # add all the inner hops to the blacklist
                        logger.error(
                            "   Unknown next peer somewhere in route.")
                        inner_hops = r.channel_hops[1:-1]
                        for i_hop, hop in enumerate(inner_hops):
                            failed_channel_source = r.node_hops[i_hop]
                            failed_channel_target = r.node_hops[i_hop + 1]
                            self.router.channel_rater.add_bad_channel(
                                hop, failed_channel_source,
                                failed_channel_target)
                    continue
                else:
                    logger.debug(result.payment_preimage)
                    logger.info("Success!\n")
                    return r.total_fee_msat
            else:  # dry
                raise DryRun

    def _get_rebalance_candidates(self, channel_id, local_balance_change,
                                  allow_unbalancing=False, strategy=None):
        """
        Determines channels, which can be used to rebalance a channel.

        If the local_balance_change is negative, the local balance of the to
        be balanced channel is tried to be reduced. This method determines
        channels with which we can rebalance by ideally balancing also the
        counterparty. However, this is not always possible  so one can also
        specify to allow unbalancing until an unbalancedness of
        UNBALANCED_CHANNEL and no more. One can also specify a strategy,
        which determines the order of channels of the rebalancing process.

        :param channel_id: the channel id of the to be rebalanced channel
        :type channel_id: int

        :param local_balance_change: amount by which the local balance of
            channel should change in sat
        :type local_balance_change: int

        :param allow_unbalancing: if unbalancing of channels should be allowed
        :type allow_unbalancing: bool

        :param strategy:
            None: By default, counterparty channels are sorted such that the
                  ones unbalanced in the opposite direction are chosen first,
                  such that they also get balanced. After them also the other
                  channels, unbalanced in the non-ideal direction are tried
                  (if allowed by allow_unbalancing).
            'feerate': Channels are sorted by increasing peer fee rate.
            'affordable': Channels are sorted by the affordable amount.
        :type strategy: str

        :return: list of channels
        :rtype: list
        """
        rebalance_candidates = []

        # select the proper lower bound of unbalancedness we allow for
        # the rebalance candidate channel (allows for unbalancing a channel)
        if allow_unbalancing:
            # target is a bit into the non-ideal direction
            lower_bound = -settings.UNBALANCED_CHANNEL
        else:
            # target is perfect balance
            lower_bound = 0
        # TODO: allow for complete depletion

        # determine the direction we use the rebalance candidate for:
        # -1: rebalance channel is sending, 1: rebalance channel is receiving
        direction = -math.copysign(1, local_balance_change)

        # logic to shift the bounds accordingly into the different
        # rebalancing directions
        for k, c in self.channel_list.items():
            if direction * c['unbalancedness'] > lower_bound:
                if allow_unbalancing:
                    c['amt_affordable'] = int(
                        c['amt_to_balanced'] +
                        direction * settings.UNBALANCED_CHANNEL
                        * c['capacity'] / 2)
                else:
                    c['amt_affordable'] = c['amt_to_balanced']
                rebalance_candidates.append(c)

        # filter channels, which can't afford a rebalance
        rebalance_candidates = [
            c for c in rebalance_candidates if not c['amt_affordable'] == 0]

        # filters by max_effective_fee_rate, as this is the minimal fee rate
        # to be paid
        rebalance_candidates = [
            c for c in rebalance_candidates
            if self._effective_fee_rate(
                c['amt_affordable'], c['peer_base_fee'], c['peer_fee_rate'])
            < self.max_effective_fee_rate]

        # need to make sure we don't rebalance with the same channel
        rebalance_candidates = [
            c for c in rebalance_candidates if c['chan_id'] != channel_id]

        # need to remove multiply connected nodes, if the counterparty channel
        # should receive (can't control last hop)
        if local_balance_change < 0:
            rebalance_candidates = [
                c for c in rebalance_candidates
                if not self._node_is_multiple_connected(c['remote_pubkey'])]

        if strategy == 'most-affordable-first':
            rebalance_candidates.sort(
                key=lambda x: direction * x['amt_affordable'], reverse=True)
        elif strategy == 'lowest-feerate-first':
            rebalance_candidates.sort(
                key=lambda x: x['peer_fee_rate'])
        elif strategy == 'match-unbalanced':
            rebalance_candidates.sort(
                key=lambda x: -direction * x['unbalancedness'])
        else:
            rebalance_candidates.sort(
                key=lambda x: direction * x['amt_to_balanced'], reverse=True)
        # TODO: for each rebalance candidate calculate the shortest path
        #  with absolute fees and sort

        return rebalance_candidates

    @staticmethod
    def _effective_fee_rate(amt_sat, base, rate):
        """
        Calculates the effective fee rate: (base_fee + fee_rate * amt) / amt

        :param amt_sat: amount in sat
        :type amt_sat: int

        :param base: base fee in sat
        :type base: float

        :param rate: fee rate
        :type rate: float

        :return: effective fee rate
        :rtype: float
        """
        amt_msat = amt_sat * 1000
        assert not (amt_msat == 0)
        rate = (base + rate * amt_msat / 1000000) / amt_msat
        return rate

    @staticmethod
    def _print_rebalance_candidates(rebalance_candidates):
        """
        Prints rebalance candidates.

        :param rebalance_candidates:
        :type rebalance_candidates: list[dict]
        """
        logger.debug(f"-------- Description --------")
        logger.debug(
            "cid: channel id\n"
            "ub: unbalancedness (see --help)\n"
            "atb: amount to be balanced [sat]\n"
            "aaf: amount affordable [sat]\n"
            "l: local balance [sat]\n"
            "r: remote balance [sat]\n"
            "bf: peer base fee [msat]\n"
            "fr: peer fee rate\n"
            "a: alias"
        )

        logger.debug(f"-------- Candidates in order of rebalance attempts "
                     f"--------")
        for c in rebalance_candidates:
            logger.debug(
                f"cid:{c['chan_id']} "
                f"ub:{c['unbalancedness']: 4.2f} "
                f"atb:{c['amt_to_balanced']: 9d} "
                f"aaf:{c['amt_affordable']: 9d} "
                f"l:{c['local_balance']: 9d} "
                f"r:{c['remote_balance']: 9d} "
                f"bf:{c['peer_base_fee']: 6d} "
                f"fr:{c['peer_fee_rate']/1E6: 1.6f} "
                f"a:{c['alias']}")

    def _extract_channel_info(self, chan_id):
        """
        Gets the channel info (policy, capacity, nodes) from the graph.
        :param chan_id: channel id
        :type chan_id: int

        :return: channel information
        :rtype: dict
        """
        # TODO: make more pythonic
        channel_info = None
        for k, c in self.channel_list.items():
            if c['chan_id'] == chan_id:
                channel_info = c
        if channel_info is None:
            raise KeyError("Channel not found (already closed?)")
        return channel_info

    @staticmethod
    def _get_source_and_target_channels(channel_one, channel_two,
                                        rebalance_direction):
        """
        Determines what the sending and receiving channel ids are.

        :param channel_one: first channel
        :type channel_one: int

        :param channel_two: second channel
        :type channel_two: int

        :param rebalance_direction: positive, if receiving, negative if sending
        :type rebalance_direction: float

        :return: sending and receiving channel
        :rtype: (int, int)
        """
        if rebalance_direction < 0:
            source = channel_one
            target = channel_two
        else:
            source = channel_two
            target = channel_one

        return source, target

    @staticmethod
    def _maximal_local_balance_change(unbalancedness_target,
                                      unbalanced_channel_info):
        """
        Tries to find out the amount to maximally send/receive given the
        relative target and channel reserve constraints for channel balance
        candidates.

        The target is expressed as a relative quantity between -1 and 1:
        -1: channel has only local balance
        0: 50:50 balanced
        1: channel has only remote balance

        :param unbalancedness_target:
            interpreted in terms of unbalancedness [-1...1]
        :type unbalancedness_target: float

        :param unbalanced_channel_info: fees, capacity, initiator info
        :type unbalanced_channel_info: dict

        :return: positive or negative amount in sat (encodes
            the decrease/increase in the local balance)
        :rtype: int
        """
        # both parties need to maintain a channel reserve of 1%
        # according to BOLT 2
        channel_reserve = int(0.01 * unbalanced_channel_info['capacity'])

        if unbalancedness_target:
            # a commit fee needs to be only respected by the channel initiator
            commit_fee = 0 if not unbalanced_channel_info['initiator'] \
                else unbalanced_channel_info['commit_fee']

            # first naively calculate the local balance change to
            # fulfill the requested target
            local_balance_target = int(
                unbalanced_channel_info['capacity'] * 0.5 *
                (-unbalancedness_target + 1.0) - commit_fee)
            local_balance_change = local_balance_target - \
                unbalanced_channel_info['local_balance']

            # TODO: clarify exact definitions of dust and htlc_cost
            # related: https://github.com/lightningnetwork/lnd/issues/1076
            # https://github.com/lightningnetwork/lightning-rfc/blob/master/03-transactions.md#fees

            dust = 700
            htlc_weight = 172
            number_htlcs = 2
            htlc_cost = int(
                number_htlcs * htlc_weight *
                unbalanced_channel_info['fee_per_kw'] / 1000)
            logger.debug(f">>> Assuming a dust limit of {dust} sat and an "
                         f"HTLC cost of {htlc_cost} sat.")

            # we can only send the local balance less the channel reserve
            # (if above the dust limit), less the cost to enforce the HTLC
            can_send = max(0, unbalanced_channel_info['local_balance']
                           - max(dust, channel_reserve) - htlc_cost - 1)

            # we can only receive the remote balance less the channel reserve
            # (if above the dust limit)
            can_receive = max(0, unbalanced_channel_info['remote_balance']
                              - max(dust, channel_reserve) - 1)

            logger.debug(f">>> Channel can send {can_send} sat and receive "
                         f"{can_receive} sat.")

            # check that we respect and enforce the channel reserve
            if (local_balance_change > 0 and
                    abs(local_balance_change)) > can_receive:
                local_balance_change = can_receive
            if (local_balance_change < 0 and
                    abs(local_balance_change)) > can_send:
                local_balance_change = -can_send

            amt_target_original = int(local_balance_change)
        else:
            # use the already calculated optimal amount for 50:50 balancedness
            amt_target_original = unbalanced_channel_info['amt_to_balanced']

        return amt_target_original

    def _node_is_multiple_connected(self, pub_key):
        """
        Checks if the node is connected to us via several channels.

        :param pub_key: node public key
        :type pub_key: str

        :return: true if number of channels to the node is larger than 1
        :rtype: bool
        """
        n_channels = 0
        for pk, chan_info in self.channel_list.items():
            if chan_info['remote_pubkey'] == pub_key:
                n_channels += 1
        if n_channels > 1:
            return True
        else:
            return False

    def rebalance(self, channel_id, dry=False, chunksize=1.0, target=None,
                  allow_unbalancing=False, strategy=None):
        """
        Automatically rebalances a selected channel with a fee cap of
        self.budget_sat and self.max_effective_fee_rate.

        Rebalancing candidates are selected among all channels which are
        unbalanced in the other direction.
        Uses :func:`self.rebalance_two_channels` for rebalancing a pairs
        of channels.

        At the moment, rebalancing channels are tried one at a time,
        so it is not yet optimized for the lowest possible fees.

        The chunksize allows for partitioning of the individual rebalancing
        attempts into smaller pieces than would maximally possible. The smaller
        the chunksize the higher is the success rate, but the rebalancing
        cost increases.

        :param channel_id:
        :type channel_id: int

        :param dry: if set, then there's a dry run
        :type dry: bool

        :param chunksize: a number between 0 and 1
        :type chunksize: float

        :param target: specifies unbalancedness after rebalancing in [-1, 1]
        :type target: float

        :param allow_unbalancing: allows counterparty channels
            to get a little bit unbalanced
        :type allow_unbalancing: bool

        :param strategy: lets you select a strategy for rebalancing order
        :type strategy: str

        :return: fees in msat paid for rebalancing
        :rtype: int

        :raises MultichannelInboundRebalanceFailure: can't rebalance channels
            with a node we are mutiply connected
        :raises NoRebalanceCandidates: there are no counterparty rebalance
            candidates
        :raises RebalanceCandidatesExhausted: no more conterparty rebalance
            candidates
        :raises TooExpensive: the rebalance became too expensive
        """
        if not (0.0 <= chunksize <= 1.0):
            raise ValueError("Chunk size must be between 0.0 and 1.0.")
        if not (-1.0 <= target <= 1.0):
            raise ValueError("Target must be between -1.0 and 1.0.")

        logger.info(f">>> Trying to rebalance channel {channel_id} "
                    f"with a max rate of {self.max_effective_fee_rate} "
                    f"and a max fee of {self.budget_sat} sat.")

        if dry:
            logger.info(f">>> This is a dry run, nothing to fear.")

        # get a fresh channel list
        self.channel_list = self.node.get_unbalanced_channels()
        unbalanced_channel_info = self.channel_list[channel_id]

        # if a target is given and it is set close to -1 or 1,
        # then we need to think about the channel reserve
        initial_local_balance_change = self._maximal_local_balance_change(
            target, unbalanced_channel_info)

        if initial_local_balance_change == 0:
            logger.info(f"Channel already balanced.")
            return 0

        # copy the original target
        local_balance_change_left = initial_local_balance_change

        # if chunksize is set, rebalance in portions of chunked_amount
        chunked_amount = int(initial_local_balance_change * chunksize)
        logger.info(f">>> Chunk size is set to {chunksize}. "
                    f"Results in chunksize of {chunked_amount} sat.")

        # determine rebalance direction 1: receive, -1: send (of channel_id)
        rebalance_direction = math.copysign(1, initial_local_balance_change)

        logger.info(
            f">>> The channel status before rebalancing is "
            f"lb:{unbalanced_channel_info['local_balance']} sat "
            f"rb:{unbalanced_channel_info['remote_balance']} sat "
            f"cap:{unbalanced_channel_info['capacity']} sat.")
        logger.debug(
            f">>> Commit fee {unbalanced_channel_info['commit_fee']} "
            f"sat. We opened channel: {unbalanced_channel_info['initiator']}. "
            f"Channel reserve: "
            f"{int(unbalanced_channel_info['capacity'] * 0.01)} sat.")
        logger.debug(
            f">>> The change in local balance towards the requested "
            f"target (ub={target if target else 0.0:3.2f})"
            f" is {initial_local_balance_change} sat.")

        # a commit fee is only accounted for, if we opened the channel
        commit_fee = 0
        if unbalanced_channel_info['initiator']:
            commit_fee = unbalanced_channel_info['commit_fee']

        expected_target = -2 * ((
            unbalanced_channel_info['local_balance'] +
            initial_local_balance_change + commit_fee) /
            float(unbalanced_channel_info['capacity'])) + 1

        logger.info(
            f">>> Trying to change the local balance by "
            f"{initial_local_balance_change} sat.\n"
            f"    The expected unbalancedness target is "
            f"{expected_target:3.2f} (respecting channel reserve), "
            f"requested target is {0 if not target else target:3.2f}.")

        node_is_multiple_connected = self._node_is_multiple_connected(
            unbalanced_channel_info['remote_pubkey'])

        if initial_local_balance_change > 0 and node_is_multiple_connected:
            # TODO: this might be too strict, figure out exact behavior
            raise MultichannelInboundRebalanceFailure(
                "Receiving-rebalancing of multiple "
                "connected node channel not supported.\n"
                "The reason is that the last hop "
                "(channel) can't be controlled by us.\n"
                "See https://github.com/lightningnetwork/"
                "lnd/issues/2966 and \n"
                "https://github.com/lightningnetwork/"
                "lightning-rfc/blob/master/"
                "04-onion-routing.md#non-strict-forwarding.\n"
                "Tip: keep only the best channel to "
                "the node and then rebalance.")

        rebalance_candidates = self._get_rebalance_candidates(
            channel_id, local_balance_change_left,
            allow_unbalancing=allow_unbalancing, strategy=strategy)
        if len(rebalance_candidates) == 0:
            raise NoRebalanceCandidates(
                "Didn't find counterparty rebalance candidates.")

        logger.info(
            f">>> There are {len(rebalance_candidates)} channels with which "
            f"we can rebalance (look at logs).")

        self._print_rebalance_candidates(rebalance_candidates)

        logger.info(
            f">>> We will try to rebalance with them one after another.")
        logger.info(
            f">>> NOTE: only individual rebalance requests "
            f"are optimized for fees:\n"
            f"    this means that there can be rebalances with "
            f"less fees afterwards,\n"
            f"    so take a look at the dry runs first, "
            f"i.e. without the --reckless flag,\n"
            f"    and set --max-fee-sat and --max-fee-rate accordingly.\n"
            f"    You may also specify a rebalancing strategy "
            f"by the --strategy flag.")
        logger.info(
            f">>> Rebalancing can take some time. Please be patient!\n")

        total_fees_msat = 0

        # loop over the rebalancing candidates
        for c in rebalance_candidates:

            if total_fees_msat >= self.budget_sat * 1000:
                raise TooExpensive(
                    f"Fee budget exhausted. "
                    f"Total fees {total_fees_msat} msat.")

            source_channel, target_channel = \
                self._get_source_and_target_channels(
                    channel_id, c['chan_id'], rebalance_direction)

            # counterparty channel affords less than total rebalance amount
            if abs(local_balance_change_left) > abs(c['amt_affordable']):
                sign = math.copysign(1, c['amt_affordable'])
                minimal_amount = sign * min(
                    abs(chunked_amount), abs(c['amt_affordable']))
                amt = -int(rebalance_direction * minimal_amount)
            # counterparty channel affords more than total rebalance amount
            else:
                sign = math.copysign(1, local_balance_change_left)
                minimal_amount = sign * min(
                    abs(local_balance_change_left), abs(chunked_amount))
                amt = int(rebalance_direction * minimal_amount)
            # amt must be always positive
            if amt < 0:
                raise RebalanceCandidatesExhausted(
                    f"Amount should not be negative! amt:{amt} sat")

            logger.info(
                f"-------- Rebalance from {source_channel} to {target_channel}"
                f" with {amt} sat --------")
            logger.info(
                f"Need to still change the local balance by "
                f"{local_balance_change_left} sat to reach the goal "
                f"of {initial_local_balance_change} sat. "
                f"Fees paid up to now: {total_fees_msat} msat.")

            # for each rebalance, get a new invoice
            invoice_r_hash = self.node.get_invoice(
                amt_msat=amt*1000,
                memo=f"lndmanage: Rebalance of channel {channel_id}.")

            # attempt the rebalance
            try:
                # be up to date with the blockheight, otherwise could lead
                # to cltv errors
                self.node.update_blockheight()
                total_fees_msat += self.rebalance_two_channels(
                    source_channel, target_channel, amt, invoice_r_hash,
                    self.budget_sat, dry=dry)
                local_balance_change_left -= int(rebalance_direction * amt)

                relative_amt_to_go = (local_balance_change_left /
                                      initial_local_balance_change)

                # perfect rebalancing is not always possible,
                # so terminate if at least 90% of amount was reached
                if relative_amt_to_go <= 0.10:
                    logger.info(
                        f"Goal is reached. Rebalancing done. "
                        f"Total fees were {total_fees_msat} msat.")
                    return total_fees_msat
            except NoRoute:
                logger.error(
                    "There was no reliable route with enough capacity.\n")
            except DryRun:
                logger.info(
                    "Would have tried this route now, but it was a dry run.\n")
            except TooExpensive:
                logger.error(
                    "Too expensive, check --max-fee-rate and --max-fee-sat.\n")
            except RebalanceFailure:
                logger.error(
                    "Failed to rebalance with this channel.\n")

        raise RebalanceCandidatesExhausted(
            "There are no further counterparty rebalance channel candidates "
            "for this channel.\n")
